/* CSS for StaffGradedAssignmentXBlock */

.sga-block .staff-modal {
    position: fixed;
    z-index:100;
    display: none;
    width: 80%;
    height: 75%;
    overflow: hidden;

    left: 10% !important; /* important is necessary to override modal JS styling */
    margin-left: 0px !important; /* important is necessary to override modal JS styling */
}

.sga-block .staff-modal .inner-wrapper {
    height: 100%;
    overflow-y: scroll;
    box-sizing: border-box;
}

.sga-block .staff-modal .not-finalized td {
    background-color: #f1f1f1
}

.sga-block table.gridtable {
    font-family: verdana,arial,sans-serif;
    font-size:11px;
    color:#333333;
    border-width: 1px;
    border-color: #666666;
    border-collapse: collapse;
}

.sga-block table.gridtable th {
    border-width: 1px;
    padding: 8px;
    border-style: solid;
    border-color: #666666;
    background-color: #dedede;
}

.sga-block table.gridtable td {
    border-width: 1px;
    padding: 8px;
    border-style: solid;
    border-color: #666666;
    background-color: #ffffff;
}

.sga-block .upload {
    position: relative;
    height: 40px;
}

/* Default edx button style */
.sga-block .upload label, .sga-block a.button {
    border: 1px solid #d2c9c9;
    border-radius: 3px;
    box-shadow: inset 0 1px 0 0 #fff;
    color: #333;
    display: inline-block;
    font-weight: bold;
    background-color: #f1f1f1;
    background-image: -webkit-linear-gradient(#f1f1f1, #d9d1d1);
    background-image: linear-gradient(#f1f1f1, #d9d1d1);
    padding: 7px 18px;
    text-decoration: none;
    text-shadow: 0 1px 0 #fcfbfb;
    background-clip: padding-box;
    white-space: nowrap;
    font-size: 0.8125em;
}

.sga-block a.button {
    -moz-user-select: none;
    box-sizing: border-box;
    line-height: normal;
    text-align: center;
    vertical-align: middle;
    white-space: nowrap;
    margin-top: 1px;
    cursor: pointer;
}

.sga-block .upload .fileupload {
    display: none;
}

.sga-block td .upload label {
    cursor: pointer;
    line-height: normal;
}

.sga-block .error {
    color: red;
}

.sga-block .display_name {
    color: inherit;
}

.sga-block table.tablesorter thead tr .header {
 	cursor: pointer;
}

.sga-block table.tablesorter thead tr .headerSortDown, table.tablesorter thead tr .headerSortUp {
    background-color: #8dbdd8;
}

.sga-block .submissions-download-link >  p {
  color: #0075b4;
}

.sga-block .solution::before {
  content: "Solution: ";
  font-weight: bold;
}

.staff-modal .task-message {
  font-family:'Open Sans', sans-serif;
  color: #4c4c4c;
  padding-bottom: 5px;
}

.staff-modal .ready-msg {
  color: darkgreen;
}

.staff-modal .preparing-msg {
  color: red;
}
